@@ -1,10 +1,29 @@
 # Copyright IBM Corp. 2021. All Rights Reserved.
 
-from typing import Optional, Tuple
+import re
+from typing import Optional, Tuple, Set
 
 from attr import attrs, evolve
+import urllib.parse
 
 
+_segment_regex: re.Pattern = re.compile(r'[a-zA-Z0-9:._+-]*')
+_segment_allowed_characters: Set[str] = set(':._+-')
+
+def _validate_segment(segment: str) -> None:
+    if _segment_regex.fullmatch(segment):
+        return
+
+    forbidden_chars = set()
+    for ch in segment:
+        if ch in forbidden_chars:
+            continue
+        if not _segment_regex.fullmatch(ch):
+            forbidden_chars.add(ch)
+
+    forbidden_chars_str = ', '.join([f"'{ch}'" for ch in forbidden_chars])
+    raise TypeError(f"segment '{segment}' contains forbidden characters: {forbidden_chars_str}")
+
 @attrs(auto_attribs=True, frozen=True, kw_only=True)
 class CpdScope:
     has_prefix: bool
@@ -20,47 +39,67 @@ class CpdScope:
 
     @classmethod
     def try_from_string(cls, s: str) -> Tuple[Optional['CpdScope'], str]:
-        if s.startswith("cpd://") or s.startswith("/"):
-            return cls._from_string_and_rest(s)
+        s = s.strip()
+        result = urllib.parse.urlparse(s)
+        if result.scheme == 'cpd' or (result.scheme == '' and result.path.startswith('/')):
+            return cls._from_parse_result(result)
         else:
             return None, s
 
     @classmethod
     def from_string(cls, s: str) -> 'CpdScope':
-        return cls._from_string_and_rest(s)[0]
+        result = urllib.parse.urlparse(s)
+        return cls._from_parse_result(result)[0]
 
     @classmethod
-    def _from_string_and_rest(cls, s: str) -> Tuple['CpdScope', str]:
+    def _from_parse_result(cls, result: urllib.parse) -> Tuple['CpdScope', str]:
         has_prefix = False
-        if s.startswith("cpd://"):
+        if result.scheme == 'cpd':
             has_prefix = True
-            s = s[len("cpd://"):]
 
-        parts = s.split("/")
-        if len(parts) < 3:
-            raise TypeError(f"Wrong format of CPD Scope: '{s}'")
+        context = result.netloc
+        if context == '':
+            context = None
 
-        context = parts[0]
-        scope_type = parts[1]
-        scope_id = parts[2]
-        other_parts = parts[3:]
+        query = result.query
+        if query == '':
+            query = None
+
+        parts = result.path.split("/")
+        if parts[0] == '':
+            del parts[0]
+
+        if len(parts) < 2:
+            if len(parts) == 1 and parts[0].find("?") != -1:
+                raise TypeError(f"CPD Scope for CPD Orchestration must use id, not query")
+            raise TypeError(f"Wrong format of CPD Scope: '{result.geturl()}'")
+
+        for part in parts:
+            _validate_segment(part)
+
+        scope_type = parts[0]
+        scope_id = parts[1]
+        other_parts = parts[2:]
 
         scope_types = [
             "projects",
             "spaces",
             "catalogs",
         ]
-        if scope_type.find("?") != -1:
-            raise TypeError(f"CPD Scope must use id, not name or tags")
+
         if scope_type not in scope_types:
             raise TypeError(f"Unknown scope type: '{scope_type}'")
 
+        remaining = "/".join(other_parts)
+        if len(other_parts) > 0 and query is not None:
+            remaining += '?' + query
+
         return CpdScope(
             has_prefix = has_prefix,
             context = context,
             scope_type = scope_type,
             scope_id = scope_id,
-        ), "/".join(other_parts)
+        ), remaining
 
 @attrs(auto_attribs=True, frozen=True, kw_only=True)
 class CpdPath:
@@ -81,24 +120,30 @@ class CpdPath:
 
     @classmethod
     def from_string(cls, s: str) -> 'CpdPath':
+        s = s.strip()
         scope, s = CpdScope.try_from_string(s)
 
+        result = urllib.parse.urlparse(s)
+        path = result.path
+        parts = path.split("/")
+
+        for part in parts:
+            _validate_segment(part)
+
+        query = result.query
+        if query == '':
+            query = None
+
         # if no scope, an additional case of "just by-id asset" must be checked
         if scope is None:
-            if s.find("/") == -1: # no slashes --- it is by-id asset
-                s = f"assets/{s}"
+            if path.find("/") == -1: # no slashes --- it is by-id asset
+                path = f"assets/{path}"
 
-        if s.count("?") > 1:
+        if path.count("?") > 0:
             raise TypeError("Question mark can only be used once.")
 
-        try:
-            qmark_idx = s.rindex("?")
-            s, query = s[:qmark_idx-1], s[qmark_idx+1:]
-        except ValueError:
-            query = None
-
-        path = CpdPath(scope = scope, path = s, query = query)
-        return path
+        result = CpdPath(scope = scope, path = path, query = query)
+        return result
 
     def resource_id(self) -> Optional[str]:
         if self.query is not None:
